/*
 * PhotoStepper -- https://www.0x50.de/rschaten/photostepper
 *
 * PhotoStepper © 2022 by Ronald Schaten is licensed under CC BY-NC 4.0.
 * To view a copy of this license, visit http://creativecommons.org/licenses/by-nc/4.0/
 *
 */

// this project makes heavy use of the following libraries:
// - https://www.arduino.cc/reference/en/libraries/arduinomenu-library/
// - https://www.arduino.cc/reference/en/libraries/stepperdriver/
// - https://www.arduino.cc/reference/en/libraries/analogkeypad/

#include <Arduino.h>
#include <AnalogKeypad.h>
#include <BasicStepperDriver.h>
#include <menu.h>
#include <menuIO/liquidCrystalOut.h>
#include <menuIO/serialIO.h>

// analog keypad
#define PIN_BUTTON 0
#define KEY_RIGHT 0
#define KEY_UP 1
#define KEY_DOWN 2
#define KEY_LEFT 3
#define KEY_SELECT 4
// 2x16 LCD
#define PIN_LCD_DB4 4
#define PIN_LCD_DB5 5
#define PIN_LCD_DB6 6
#define PIN_LCD_DB7 7
#define PIN_LCD_RS 8
#define PIN_LCD_EN 9
#define PIN_LCD_BL 10
// stepper motor
#define PIN_STEPPER_DIR 2
#define PIN_STEPPER_STEP 3
// on-board led
#define PIN_LED 13

// motor configuration
#define STEPPER_STEPS 200
#define STEPPER_RPM 120
#define STEPPER_MICROSTEPS 1
#define STEPPER_ACCEL 2000
#define STEPPER_DECEL 1000

#define MAX_DEPTH 2

const int KeypadMap[] = {0, 96, 250, 402, 636};
const uint16_t KeypadHoldTimeMs = 1000;
const uint8_t KeypadAnalogPin = PIN_BUTTON;
AnalogKeypad keypad(KeypadAnalogPin, KeypadMap, countof(KeypadMap), KeypadHoldTimeMs);

LiquidCrystal lcd(PIN_LCD_RS, PIN_LCD_EN, PIN_LCD_DB4, PIN_LCD_DB5, PIN_LCD_DB6, PIN_LCD_DB7);

BasicStepperDriver stepper(STEPPER_STEPS, PIN_STEPPER_DIR, PIN_STEPPER_STEP);

enum State { STATE_MENU, STATE_JOG, STATE_RUN };
State state = STATE_MENU;

void lcd_print(int line, String string) {
  Serial.print("lcd_print(");
  Serial.print(line);
  Serial.print(", \"");
  Serial.print(string);
  Serial.println("\")");
  lcd.setCursor(0, line);
  lcd.print(string);
  for (int i = string.length(); i<16; i++) lcd.print(' ');
}

// ===

result enter_jog() {
  Serial.println("enter_jog()");
  lcd_print(0, "Jogging...");
  lcd_print(1, "[sel] to exit");
  state = STATE_JOG;
  return proceed;
}

MENU(mainMenu, "PhotoStepper", doNothing, noEvent, wrapStyle
  ,OP("Jog", enter_jog, enterEvent)
  ,OP("Sub1", doNothing, noEvent)
  ,OP("Sub2", doNothing, noEvent)
);

MENU_OUTPUTS(out, MAX_DEPTH
  ,LIQUIDCRYSTAL_OUT(lcd, {0, 0, 16, 2})
  ,NONE
);

serialIn in(Serial);
NAVROOT(nav, mainMenu, MAX_DEPTH, in, out);

// ===

void ButtonHandler_menu(const ButtonParam& param) {
  Serial.print("ButtonHandler_menu(");
  Serial.print(param.button);
  Serial.print(") - ");
  switch (param.state) {
    case ButtonState_Click:
      Serial.println("Click");
      switch (param.button) {
        case KEY_UP:
          nav.doNav(downCmd);
          break;
        case KEY_DOWN:
          nav.doNav(upCmd);
          break;
        case KEY_SELECT:
          nav.doNav(enterCmd);
          break;
      }
      nav.doOutput();
      break;
    default:
      Serial.println(param.state);
  }
}

void ButtonHandler_jog(const ButtonParam& param) {
  Serial.print("ButtonHandler_jog(");
  Serial.print(param.button);
  Serial.print(") - ");
  switch (param.state) {
    case ButtonState_Click:
      Serial.println("Click");
      switch (param.button) {
        case KEY_SELECT:
          state = STATE_MENU;
          mainMenu.dirty = true;
          break;
      }
      nav.doOutput();
      break;
    case ButtonState_Up:
      Serial.println("Up");
      stepper.stop();
      stepper.setRPM(STEPPER_RPM);
      break;
    case ButtonState_Down:
      Serial.println("Down");
      switch (param.button) {
        case KEY_UP:
          stepper.startMove(100 * 360);
          break;
        case KEY_DOWN:
          stepper.startMove(-100 * 360);
          break;
      }
      break;
    case ButtonState_Hold:
      Serial.println("Hold");
      stepper.setRPM(2 * STEPPER_RPM);
      switch (param.button) {
        case KEY_UP:
          stepper.startMove(100 * 360);
          break;
        case KEY_DOWN:
          stepper.startMove(-100 * 360);
          break;
      }
      break;
    default:
      Serial.println();
  }
}

void ButtonHandler_run(const ButtonParam& param) {
  Serial.print("ButtonHandler_run(");
  Serial.print(param.button);
  Serial.print(") - ");
  Serial.println("TODO");
}

void setup() {
  Serial.begin(115200);
  while(!Serial);
  Serial.println("PhotoStepper -- schatenseite.de");

  // control backlight
  pinMode(PIN_LCD_BL, OUTPUT);
  digitalWrite(PIN_LCD_BL, HIGH);

  stepper.begin(STEPPER_RPM, STEPPER_MICROSTEPS);
  stepper.setSpeedProfile(stepper.LINEAR_SPEED, STEPPER_ACCEL, STEPPER_DECEL);

  lcd.begin(16,2);
  lcd_print(0, "PhotoStepper");
  lcd_print(1, "schatenseite.de");
  delay(2000);
  nav.doOutput();
}

void loop() {
  switch (state) {
    case STATE_MENU:
      keypad.loop(ButtonHandler_menu);
      break;
    case STATE_JOG:
      keypad.loop(ButtonHandler_jog);
      break;
    case STATE_RUN:
      keypad.loop(ButtonHandler_run);
      break;
  }
  stepper.nextAction();
}
